# SPDX-FileCopyrightText: 2022-2025 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional

from pontos.models import Model


@dataclass
class AcceptanceLevel(Model):
    description: str
    last_modified: datetime


@dataclass
class Source(Model):
    last_modified: datetime
    created: datetime
    name: Optional[str] = None
    source_identifiers: List[str] = field(default_factory=list)
    contact_email: Optional[str] = None
    v2_acceptance_level: Optional[AcceptanceLevel] = None
    v3_acceptance_level: Optional[AcceptanceLevel] = None
    cwe_acceptance_level: Optional[AcceptanceLevel] = None



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

# pylint: disable=line-too-long, arguments-differ, redefined-builtin
# ruff: noqa: E501

from datetime import datetime, timezone
from typing import Any, Optional
from unittest.mock import MagicMock, patch
from uuid import UUID, uuid4

from httpx import AsyncClient, Response

from pontos.errors import PontosError
from pontos.nvd.api import now
from pontos.nvd.cpe_match.api import MAX_CPE_MATCHES_PER_PAGE, CPEMatchApi
from tests import AsyncMock, IsolatedAsyncioTestCase, aiter, anext
from tests.nvd import get_cpe_match_data


def uuid_replace_str(uuid: UUID, iteration: int, number: int) -> str:
    id_str = str(uuid).rsplit("-", 2)
    return f"{id_str[0]}-{iteration:04}-{number:012}"


def uuid_replace(uuid: UUID, iteration: int, number: int) -> UUID:
    return UUID(uuid_replace_str(uuid, iteration, number))


def generate_cpe_name(iteration: int, number: int) -> str:
    return f"cpe:2.3:a:acme:test-app:1.{iteration-1}.{number-1}:*:*:*:*:*:*:*"


def create_cpe_match_response(
    match_criteria_id: UUID,
    cpe_name_id: UUID,
    *,
    update: Optional[dict[str, Any]] = None,
    results: int = 1,
    iteration: int = 1,
) -> MagicMock:
    match_strings = [
        {
            "match_string": get_cpe_match_data(
                {
                    "match_criteria_id": f"{uuid_replace_str(match_criteria_id, iteration, i)}",
                    "criteria": generate_cpe_name(iteration, i),
                    "matches": [
                        {
                            "cpe_name": generate_cpe_name(iteration, i),
                            "cpe_name_id": f"{uuid_replace_str(cpe_name_id, iteration, i)}",
                        }
                    ],
                }
            )
        }
        for i in range(1, results + 1)
    ]

    data = {
        "match_strings": match_strings,
        "results_per_page": results,
    }
    if update:
        data.update(update)

    response = MagicMock(spec=Response)
    response.json.return_value = data
    return response


def create_cpe_match_responses(
    match_criteria_id: UUID,
    cpe_name_id: UUID,
    responses: int = 2,
    results_per_response: int = 1,
) -> list[MagicMock]:
    return [
        create_cpe_match_response(
            match_criteria_id=match_criteria_id,
            cpe_name_id=cpe_name_id,
            update={"total_results": responses * results_per_response},
            results=results_per_response,
            iteration=i,
        )
        for i in range(1, responses + 1)
    ]


class CPEMatchApiTestCase(IsolatedAsyncioTestCase):
    @patch("pontos.nvd.api.time.monotonic", autospec=True)
    @patch("pontos.nvd.api.AsyncClient", spec=AsyncClient)
    def setUp(self, async_client: MagicMock, monotonic_mock: MagicMock) -> None:
        self.http_client = AsyncMock()
        async_client.return_value = self.http_client
        monotonic_mock.return_value = 0
        self.api = CPEMatchApi()

    async def test_no_match_criteria_id(self):
        with self.assertRaises(PontosError):
            await self.api.cpe_match(None)

    async def test_no_match_strings(self):
        data = {
            "match_strings": [],
            "results_per_page": 1,
        }
        response = MagicMock(spec=Response)
        response.json.return_value = data
        self.http_client.get.return_value = response

        with self.assertRaises(PontosError):
            await self.api.cpe_match("DOES-NOT-EXIST")

    async def test_cpe_match(self):
        match_criteria_id = uuid_replace(uuid4(), 1, 1)
        cpe_name_id = uuid_replace(uuid4(), 1, 1)

        self.http_client.get.return_value = create_cpe_match_response(
            match_criteria_id, cpe_name_id
        )

        cpe_match_string = await self.api.cpe_match(match_criteria_id)

        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={},
            params={"matchCriteriaId": str(match_criteria_id)},
        )

        self.assertEqual(
            match_criteria_id,
            cpe_match_string.match_criteria_id,
        )
        self.assertEqual(
            generate_cpe_name(1, 1),
            cpe_match_string.criteria,
        )
        self.assertEqual(
            cpe_name_id,
            cpe_match_string.matches[0].cpe_name_id,
        )
        self.assertEqual(
            generate_cpe_name(1, 1),
            cpe_match_string.matches[0].cpe_name,
        )

    @patch("pontos.nvd.cpe_match.api.now", spec=now)
    async def test_cpe_matches_last_modified_start_date(
        self, now_mock: MagicMock
    ):
        match_criteria_id = uuid4()
        cpe_name_id = uuid4()

        now_mock.return_value = datetime(2019, 8, 30, tzinfo=timezone.utc)
        self.http_client.get.side_effect = create_cpe_match_responses(
            match_criteria_id, cpe_name_id
        )

        it = aiter(
            self.api.cpe_matches(
                last_modified_start_date=datetime(
                    2019, 6, 1, tzinfo=timezone.utc
                )
            )
        )
        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 1, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={},
            params={
                "startIndex": 0,
                "lastModStartDate": "2019-06-01T00:00:00.000+00:00",
                "lastModEndDate": "2019-08-30T00:00:00.000+00:00",
                "resultsPerPage": MAX_CPE_MATCHES_PER_PAGE,
            },
        )

        self.http_client.get.reset_mock()

        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 2, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={},
            params={
                "startIndex": 1,
                "lastModStartDate": "2019-06-01T00:00:00.000+00:00",
                "lastModEndDate": "2019-08-30T00:00:00.000+00:00",
                "resultsPerPage": 1,
            },
        )

        with self.assertRaises(StopAsyncIteration):
            cpe_match = await anext(it)

    @patch("pontos.nvd.cpe_match.api.now", spec=now)
    async def test_cpe_matches_last_modified_end_date(
        self, now_mock: MagicMock
    ):
        match_criteria_id = uuid4()
        cpe_name_id = uuid4()

        now_mock.return_value = datetime(2019, 8, 30, tzinfo=timezone.utc)
        self.http_client.get.side_effect = create_cpe_match_responses(
            match_criteria_id, cpe_name_id
        )

        it = aiter(
            self.api.cpe_matches(
                last_modified_end_date=datetime(2019, 8, 1, tzinfo=timezone.utc)
            )
        )
        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 1, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={},
            params={
                "startIndex": 0,
                "lastModEndDate": "2019-08-01T00:00:00.000+00:00",
                "resultsPerPage": MAX_CPE_MATCHES_PER_PAGE,
            },
        )

        self.http_client.get.reset_mock()

        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 2, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={},
            params={
                "startIndex": 1,
                "lastModEndDate": "2019-08-01T00:00:00.000+00:00",
                "resultsPerPage": 1,
            },
        )

        with self.assertRaises(StopAsyncIteration):
            cpe_match = await anext(it)

    async def test_cpe_matches_cve_id(self):
        match_criteria_id = uuid4()
        cpe_name_id = uuid4()

        self.http_client.get.side_effect = create_cpe_match_responses(
            match_criteria_id, cpe_name_id
        )

        it = aiter(self.api.cpe_matches(cve_id="CVE-2010-3574"))
        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 1, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={},
            params={
                "startIndex": 0,
                "cveId": "CVE-2010-3574",
                "resultsPerPage": MAX_CPE_MATCHES_PER_PAGE,
            },
        )

        self.http_client.get.reset_mock()

        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 2, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={},
            params={
                "startIndex": 1,
                "cveId": "CVE-2010-3574",
                "resultsPerPage": 1,
            },
        )

        with self.assertRaises(StopAsyncIteration):
            cpe_match = await anext(it)

    async def test_cpe_matches_match_string_search(self):
        match_criteria_id = uuid4()
        cpe_name_id = uuid4()

        self.http_client.get.side_effect = create_cpe_match_responses(
            match_criteria_id, cpe_name_id
        )

        it = aiter(
            self.api.cpe_matches(
                match_string_search="cpe:2.3:a:sun:jre:*:*:*:*:*:*:*:*"
            )
        )
        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 1, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={},
            params={
                "startIndex": 0,
                "matchStringSearch": "cpe:2.3:a:sun:jre:*:*:*:*:*:*:*:*",
                "resultsPerPage": MAX_CPE_MATCHES_PER_PAGE,
            },
        )

        self.http_client.get.reset_mock()

        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 2, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={},
            params={
                "startIndex": 1,
                "matchStringSearch": "cpe:2.3:a:sun:jre:*:*:*:*:*:*:*:*",
                "resultsPerPage": 1,
            },
        )

        with self.assertRaises(StopAsyncIteration):
            cpe_match = await anext(it)

    async def test_cpe_matches_request_results(self):
        match_criteria_id = uuid4()
        cpe_name_id = uuid4()

        self.http_client.get.side_effect = create_cpe_match_responses(
            match_criteria_id=match_criteria_id,
            cpe_name_id=cpe_name_id,
            results_per_response=2,
        )

        it = aiter(self.api.cpe_matches(request_results=10))
        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 1, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={},
            params={
                "startIndex": 0,
                "resultsPerPage": 10,
            },
        )

        self.http_client.get.reset_mock()
        cpe_match = await anext(it)
        self.assertEqual(
            uuid_replace(match_criteria_id, 1, 2), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_not_called()

        self.http_client.get.reset_mock()

        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 2, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={},
            params={
                "startIndex": 2,
                "resultsPerPage": 2,
            },
        )

        self.http_client.get.reset_mock()
        cpe_match = await anext(it)
        self.assertEqual(
            uuid_replace(match_criteria_id, 2, 2), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_not_called()

        with self.assertRaises(Exception):
            cpe_match = await anext(it)

    async def test_cpe_match_caching(self):
        match_criteria_id = uuid4()
        cpe_name_id = uuid4()

        responses = create_cpe_match_responses(
            match_criteria_id=match_criteria_id,
            cpe_name_id=cpe_name_id,
            results_per_response=3,
        )
        self.http_client.get.side_effect = responses
        response_matches = [
            [
                match_string["match_string"]["matches"]
                for match_string in response.json.return_value["match_strings"]
            ]
            for response in responses
        ]

        # Make matches of first match_string identical in each response
        response_matches[1][0][0]["cpe_name"] = response_matches[0][0][0][
            "cpe_name"
        ]
        response_matches[1][0][0]["cpe_name_id"] = response_matches[0][0][0][
            "cpe_name_id"
        ]
        # Make matches of second match_string only have the same cpe_name_id
        response_matches[1][1][0]["cpe_name_id"] = response_matches[0][1][0][
            "cpe_name_id"
        ]
        # Leave matches of third match_string different from each other

        it = aiter(self.api.cpe_matches(request_results=10))
        received = [item async for item in it]

        # First matches in each response of three items must be identical objects
        self.assertIs(received[0].matches[0], received[3].matches[0])

        # Second matches in each response of three items must only have same cpe_name_id
        self.assertIsNot(received[1].matches[0], received[4].matches[0])
        self.assertEqual(
            received[1].matches[0].cpe_name_id,
            received[4].matches[0].cpe_name_id,
        )
        self.assertNotEqual(
            received[1].matches[0].cpe_name, received[4].matches[0].cpe_name
        )

        # Third matches in each response of three items must be different
        self.assertIsNot(received[2].matches[0], received[5].matches[0])
        self.assertNotEqual(
            received[2].matches[0].cpe_name_id,
            received[5].matches[0].cpe_name_id,
        )
        self.assertNotEqual(
            received[2].matches[0].cpe_name, received[5].matches[0].cpe_name
        )

    async def test_context_manager(self):
        async with self.api:
            pass

        self.http_client.__aenter__.assert_awaited_once()
        self.http_client.__aexit__.assert_awaited_once()


class CPEMatchApiWithTokenTestCase(IsolatedAsyncioTestCase):
    @patch("pontos.nvd.api.time.monotonic", autospec=True)
    @patch("pontos.nvd.api.AsyncClient", spec=AsyncClient)
    def setUp(self, async_client: MagicMock, monotonic_mock: MagicMock) -> None:
        self.http_client = AsyncMock()
        async_client.return_value = self.http_client
        monotonic_mock.return_value = 0
        self.api = CPEMatchApi(token="token123")

    async def test_cpe_matches_request_results_with_token(self):
        match_criteria_id = uuid4()
        cpe_name_id = uuid4()

        self.http_client.get.side_effect = create_cpe_match_responses(
            match_criteria_id=match_criteria_id,
            cpe_name_id=cpe_name_id,
            results_per_response=2,
        )

        it = aiter(self.api.cpe_matches(request_results=10))
        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 1, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={"apiKey": "token123"},
            params={
                "startIndex": 0,
                "resultsPerPage": 10,
            },
        )

        self.http_client.get.reset_mock()
        cpe_match = await anext(it)
        self.assertEqual(
            uuid_replace(match_criteria_id, 1, 2), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_not_called()

        self.http_client.get.reset_mock()

        cpe_match = await anext(it)

        self.assertEqual(
            uuid_replace(match_criteria_id, 2, 1), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/cpematch/2.0",
            headers={"apiKey": "token123"},
            params={
                "startIndex": 2,
                "resultsPerPage": 2,
            },
        )

        self.http_client.get.reset_mock()
        cpe_match = await anext(it)
        self.assertEqual(
            uuid_replace(match_criteria_id, 2, 2), cpe_match.match_criteria_id
        )
        self.http_client.get.assert_not_called()

        with self.assertRaises(Exception):
            cpe_match = await anext(it)

    @patch("pontos.nvd.api.time.monotonic", autospec=True)
    @patch("pontos.nvd.api.asyncio.sleep", autospec=True)
    async def test_rate_limit_with_token(
        self,
        sleep_mock: MagicMock,
        monotonic_mock: MagicMock,
    ):
        match_criteria_id = uuid4()
        cpe_name_id = uuid4()
        self.http_client.get.side_effect = create_cpe_match_responses(
            match_criteria_id, cpe_name_id, 8
        )
        monotonic_mock.side_effect = [10, 11]

        it = aiter(self.api.cpe_matches())
        await anext(it)
        await anext(it)
        await anext(it)
        await anext(it)
        await anext(it)

        sleep_mock.assert_not_called()

        await anext(it)

        sleep_mock.assert_not_called()



# SPDX-FileCopyrightText: 2022-2025 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later

from datetime import datetime
from typing import (
    Iterable,
    Iterator,
    Optional,
)

from httpx import Timeout

from pontos.nvd.api import (
    DEFAULT_TIMEOUT_CONFIG,
    JSON,
    NVDApi,
    NVDResults,
    Params,
    format_date,
    now,
)
from pontos.nvd.models.source import Source

__all__ = ("SourceApi",)

DEFAULT_NIST_NVD_SOURCE_URL = (
    "https://services.nvd.nist.gov/rest/json/source/2.0"
)
MAX_SOURCES_PER_PAGE = 1000


def _result_iterator(data: JSON) -> Iterator[Source]:
    sources: Iterable = data.get("sources", [])  # type: ignore
    return (Source.from_dict(source) for source in sources)


class SourceApi(NVDApi):
    """
    API for querying the NIST NVD source API.

    Should be used as an async context manager.

    Example:
        .. code-block:: python

            from pontos.nvd.source import SourceApi

            async with SourceApi() as api:
                async for source in api.sources():
                    print(source)
    """

    def __init__(
        self,
        *,
        token: Optional[str] = None,
        timeout: Optional[Timeout] = DEFAULT_TIMEOUT_CONFIG,
        rate_limit: bool = True,
        request_attempts: int = 1,
    ) -> None:
        """
        Create a new instance of the source API.

        Args:
            token: The API key to use. Using an API key allows to run more
                requests at the same time.
            timeout: Timeout settings for the HTTP requests
            rate_limit: Set to False to ignore rate limits. The public rate
                limit (without an API key) is 5 requests in a rolling 30 second
                window. The rate limit with an API key is 50 requests in a
                rolling 30 second window.
                See https://nvd.nist.gov/developers/start-here#divRateLimits
                Default: True.
            request_attempts: The number of attempts per HTTP request. Defaults to 1.
        """
        super().__init__(
            DEFAULT_NIST_NVD_SOURCE_URL,
            token=token,
            timeout=timeout,
            rate_limit=rate_limit,
            request_attempts=request_attempts,
        )

    def sources(
        self,
        *,
        last_modified_start_date: Optional[datetime] = None,
        last_modified_end_date: Optional[datetime] = None,
        source_identifier: Optional[str] = None,
        request_results: Optional[int] = None,
        start_index: int = 0,
        results_per_page: Optional[int] = None,
    ) -> NVDResults[Source]:
        """
        Get all sources for the provided arguments

        https://nvd.nist.gov/developers/data-sources#divGetSource

        Args:
            last_modified_start_date: Return all sources modified after this date.
            last_modified_end_date: Return all sources modified before this date.
                If last_modified_start_date is set but no
                last_modified_end_date is passed it is set to now.
            source_identifier: Return all source records where the source identifier matches.
            request_results: Number of sources to download. Set to None
                (default) to download all available CPEs.
            start_index: Index of the first source to be returned. Useful only for
                paginated requests that should not start at the first page.
            results_per_page: Number of results in a single requests. Mostly
                useful for paginated requests.

        Returns:
            A NVDResponse for sources

        Examples:
            .. code-block:: python

                from pontos.nvd.source import SourceApi

                async with SourceApi() as api:
                    async for source in api.sources(source_identifier="cve@mitre.org"):
                        print(source)
        """
        params: Params = {}
        if last_modified_start_date:
            params["lastModStartDate"] = format_date(last_modified_start_date)
            if not last_modified_end_date:
                params["lastModEndDate"] = format_date(now())
        if last_modified_end_date:
            params["lastModEndDate"] = format_date(last_modified_end_date)

        if source_identifier:
            params["sourceIdentifier"] = source_identifier

        results_per_page = min(
            results_per_page or MAX_SOURCES_PER_PAGE,
            request_results or MAX_SOURCES_PER_PAGE,
        )
        return NVDResults(
            self,
            params,
            _result_iterator,
            request_results=request_results,
            results_per_page=results_per_page,
            start_index=start_index,
        )

    async def __aenter__(self) -> "SourceApi":
        await super().__aenter__()
        return self



# SPDX-FileCopyrightText: 2024 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later
#

from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, List, Optional
from uuid import UUID

from pontos.models import Model


@dataclass
class CPEMatch(Model):
    """
    Represents a single CPE match.

    Attributes:
        cpe_name: Name of the matching CPE
        cpe_name_id: Name ID of the matching CPE
    """

    cpe_name: str
    cpe_name_id: UUID


@dataclass
class CPEMatchString(Model):
    """
    Represents a CPE match string, matching criteria to one or more CPEs

    Attributes:
        match_criteria_id: The identifier of the CPE match
        criteria: The CPE formatted match criteria
        version_start_including: Optional start of the matching version range, including the given version
        version_start_excluding: Optional start of the matching version range, excluding the given version
        version_end_including: Optional end of the matching version range, including the given version
        version_end_excluding: Optional end of the matching version range, excluding the given version
        status: Status of the CPE match
        cpe_last_modified: Optional date the CPEs list of the match was last modified
        created: Creation date of the CPE
        last_modified: Last modification date of the CPE
        matches: List of CPEs matching the criteria string and the optional range limits
    """

    match_criteria_id: UUID
    criteria: str
    status: str
    created: datetime
    last_modified: datetime
    cpe_last_modified: Optional[datetime] = None
    matches: List[CPEMatch] = field(default_factory=list)
    version_start_including: Optional[str] = None
    version_start_excluding: Optional[str] = None
    version_end_including: Optional[str] = None
    version_end_excluding: Optional[str] = None

    @classmethod
    def from_dict_with_cache(
        cls,
        data: dict[str, Any],
        cpe_match_cache: dict[str, CPEMatch],
    ):
        """
        Create a CPEMatchString model from a dict, reusing
        duplicate CPEMatch objects to reduce memory usage if a cache
        dict is given.

        Args:
            data: The JSON dict to generate the model from
            cpe_match_cache: A dictionary to store CPE matches or None
                to not cache and reused CPE matches
        """
        new_match_string = cls.from_dict(data)

        for i, match in enumerate(new_match_string.matches):
            cached_match: Optional[CPEMatch] = cpe_match_cache.get(
                match.cpe_name_id
            )
            if cached_match and cached_match.cpe_name == match.cpe_name:
                new_match_string.matches[i] = cached_match
            else:
                cpe_match_cache[match.cpe_name_id] = match
        return new_match_string



# SPDX-FileCopyrightText: 2023 Greenbone AG
#
# SPDX-License-Identifier: GPL-3.0-or-later


from datetime import datetime, timezone
from typing import Any, Optional
from unittest.mock import MagicMock, patch

from httpx import AsyncClient, Response

from pontos.nvd.models.source import AcceptanceLevel
from pontos.nvd.source.api import MAX_SOURCES_PER_PAGE, SourceApi
from tests import AsyncMock, IsolatedAsyncioTestCase, aiter, anext
from tests.nvd import get_source_data


def create_source_response(
    name: str, update: Optional[dict[str, Any]] = None
) -> MagicMock:
    data = {
        "sources": [get_source_data({"name": name})],
        "results_per_page": 1,
    }
    if update:
        data.update(update)

    response = MagicMock(spec=Response)
    response.json.return_value = data
    return response


def create_source_responses(count: int = 2) -> list[MagicMock]:
    return [
        create_source_response(f"MITRE-{i}", {"total_results": count})
        for i in range(1, count + 1)
    ]


class SourceAPITestCase(IsolatedAsyncioTestCase):
    @patch("pontos.nvd.api.AsyncClient", spec=AsyncClient)
    def setUp(self, async_client: MagicMock) -> None:
        self.http_client = AsyncMock()
        async_client.return_value = self.http_client
        self.api = SourceApi(token="token")

    async def test_sources(self):
        self.http_client.get.side_effect = create_source_responses()

        it = aiter(self.api.sources())
        source = await anext(it)

        self.assertEqual(source.name, "MITRE-1")
        self.assertEqual(
            source.contact_email,
            "cve@mitre.org",
        )
        self.assertEqual(
            source.source_identifiers,
            [
                "cve@mitre.org",
                "8254265b-2729-46b6-b9e3-3dfca2d5bfca",
            ],
        )
        self.assertEqual(
            source.last_modified,
            datetime(2019, 9, 9, 16, 18, 45, 930000, tzinfo=timezone.utc),
        )
        self.assertEqual(
            source.created,
            datetime(2019, 9, 9, 16, 18, 45, 930000, tzinfo=timezone.utc),
        )
        self.assertIsNone(source.v2_acceptance_level)
        self.assertEqual(
            source.v3_acceptance_level,
            AcceptanceLevel(
                "Contributor",
                datetime(2025, 1, 30, 0, 0, 20, 107000, tzinfo=timezone.utc),
            ),
        )
        self.assertEqual(
            source.cwe_acceptance_level,
            AcceptanceLevel(
                "Reference",
                datetime(2025, 1, 24, 0, 0, 0, 43000, tzinfo=timezone.utc),
            ),
        )

        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/source/2.0",
            headers={"apiKey": "token"},
            params={
                "startIndex": 0,
                "resultsPerPage": MAX_SOURCES_PER_PAGE,
            },
        )

        self.http_client.get.reset_mock()

        source = await anext(it)
        self.assertEqual(source.name, "MITRE-2")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/source/2.0",
            headers={"apiKey": "token"},
            params={"startIndex": 1, "resultsPerPage": 1},
        )

        with self.assertRaises(StopAsyncIteration):
            await anext(it)

    async def test_sources_change_dates(self):
        self.http_client.get.side_effect = create_source_responses()

        it = aiter(
            self.api.sources(
                last_modified_start_date=datetime(
                    2025, 1, 1, tzinfo=timezone.utc
                ),
                last_modified_end_date=datetime(
                    2025, 1, 31, tzinfo=timezone.utc
                ),
            )
        )
        source = await anext(it)

        self.assertEqual(source.name, "MITRE-1")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/source/2.0",
            headers={"apiKey": "token"},
            params={
                "startIndex": 0,
                "lastModStartDate": "2025-01-01T00:00:00.000+00:00",
                "lastModEndDate": "2025-01-31T00:00:00.000+00:00",
                "resultsPerPage": MAX_SOURCES_PER_PAGE,
            },
        )

        self.http_client.get.reset_mock()

        source = await anext(it)

        self.assertEqual(source.name, "MITRE-2")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/source/2.0",
            headers={"apiKey": "token"},
            params={
                "startIndex": 1,
                "lastModStartDate": "2025-01-01T00:00:00.000+00:00",
                "lastModEndDate": "2025-01-31T00:00:00.000+00:00",
                "resultsPerPage": 1,
            },
        )

        with self.assertRaises(StopAsyncIteration):
            source = await anext(it)

    async def test_sources_source_identifier(self):
        self.http_client.get.side_effect = create_source_responses()

        it = aiter(self.api.sources(source_identifier="cve@mitre.org"))
        source = await anext(it)
        self.assertEqual(source.name, "MITRE-1")

        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/source/2.0",
            headers={"apiKey": "token"},
            params={
                "startIndex": 0,
                "resultsPerPage": MAX_SOURCES_PER_PAGE,
                "sourceIdentifier": "cve@mitre.org",
            },
        )

        self.http_client.get.reset_mock()

        source = await anext(it)
        self.assertEqual(source.name, "MITRE-2")
        self.http_client.get.assert_awaited_once_with(
            "https://services.nvd.nist.gov/rest/json/source/2.0",
            headers={"apiKey": "token"},
            params={
                "startIndex": 1,
                "resultsPerPage": 1,
                "sourceIdentifier": "cve@mitre.org",
            },
        )

        with self.assertRaises(StopAsyncIteration):
            await anext(it)

    async def test_context_manager(self):
        async with self.api:
            pass

        self.http_client.__aenter__.assert_awaited_once()
        self.http_client.__aexit__.assert_awaited_once()
